# --------------------------
# Controllers folder
# Class definition
#
# Async : thread management w the app
# --------------------------

""" imports: logging, config, tkinter, threading, etc """
from biaspeech.utils.config import Config
import logging
import logging.config
from threading import Thread

""" imports: MVC """
from biaspeech.models.openAI.ai import AI
from biaspeech.models.robot.robot import Bia
from biaspeech.models.speech.prompt import Prompt

""" objects: logging and config """
conf = Config()  # config


class Async(Thread):
    
    def __init__(self, text_in):
       super().__init__()
       
       """ util objects """
       self.conf = Config()  # config
       self.logger = logging.getLogger(__name__)  # logger
               
       """ attributes """  
       self.text_in = text_in 
       self.ai = AI()  # object for openAI
       self.output = ""
        
    """ run """
    def run(self): 
       self.output = ""
       self.output = self.process(self.text_in)  # process 
 
    """ process """   
    def process(self, text_in):
        prompt = Prompt(text_in)  # prompt object  
        bia = Bia(prompt, self.ai)  # bia object
        output = bia.get_output().lower()  # get the output 
       
        bia.run()  # run the action if any    
        bia.save()  # save to sqlite    
                
        return output



# --------------------------
# Models folder
# Class definition
#
# Text2Speech : define the text to speech model
# --------------------------

""" imports: logging and config """
import logging
import logging.config
from biaspeech.utils.config import Config

""" imports: pyttsx3 """
import pyttsx3 
        
""" objects:  """
conf = Config() # config
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=conf.LOGLEVEL)

class Text2Speech:    

    """ constructor """
    def __init__(self, language, rate, volume, voice_macos, voice_raspberry): 

        """ util objects """
        self.logger = logging.getLogger(__name__) # logger

        """ attributes """
        self.language = language # language
        self.rate = rate # rate
        self.volume = volume # volume
        self.voice_macos = voice_macos # voice
        self.voice_raspberry = voice_raspberry # voice
        
    """ do_speech """        
    def do_speech(self, text):
        try:          
            engine = pyttsx3.init()
                 
            engine.setProperty('language', self.language) # language
            engine.setProperty('rate', self.rate) # speech rate (words per minute), default is 200
            engine.setProperty('volume', self.volume) # volume (0.0 to 1.0), default is 1.0 
            
            if conf.OS == "macos":
                engine.setProperty('voice', self.voice_macos)  # voice          
            else:
                engine.setProperty('voice', self.voice_raspberry)  # voice      
              
            
            engine.say(text) # say some
            engine.runAndWait() # go
            
        except Exception as err:
            self.logger.warning(f"conf.ERR_D {err=}, {type(err)=}") # inform the user
            raise



# --------------------------
# Models folder
# Class definition
#
# Mia_utils : utils for mia table
# --------------------------
from websockets.version import commit

""" imports: logging and config """
import logging
import logging.config
from biaspeech.utils.config import Config

""" imports: sqlite """
import sqlite3

""" objects: logging and config """
conf = Config() # config
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=conf.LOGLEVEL)

CONN = sqlite3.connect(conf.DATABASE, check_same_thread=False) # connection to sqlite
CURSOR = CONN.cursor() # cursor

class Mia_utils:
    
    """ constructor """
    def __init__(self): 
        pass
    
    """ get_input_en """
    def get_input_en(cls, input, lang):
        sql = """    
        select input_en 
        from mia
        where input = ? and lang = ?
        order by w_neutral desc, w_positive desc, datetime desc
        limit 1
        """        
        table_rows = CURSOR.execute(sql, (input, lang)).fetchall()

        return (table_rows)  
    



# --------------------------
# Models folder
# Class definition
#
# Netflix : define the Netflix (Block) model
# --------------------------

""" imports: logging and config """
import logging
import logging.config
from biaspeech.utils.config import Config
import webbrowser

""" objects:  """
conf = Config()  # config
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=conf.LOGLEVEL)

""" MVC """
from biaspeech.models.blocks import block


class Netflix(block.Block):
    
    """ constructor """
    def __init__(self, prompt, ai): 
        block.Block.__init__(self, prompt, ai)
        
    """ get_output """        
    def get_output(self): 
        return(conf.OUTPUT_WEB)
 
    """ get_file """        
    def get_file(self):
        pass
       
    """ run """        
    def run(self, file):
        self.logger.info("run w Netflix")  # inform the user      
        
        search = self.prompt.mia.input.replace("netflix ", "")  # build the url
        url = conf.NETFLIX  # build the url
        
        if search != "":  # build the url
            url = url + "search?q=" + search
        
        webbrowser.open(url)  # open-up the url
     



# --------------------------
# Models folder
# Class definition
#
# Bia : define the Bia model
# --------------------------

""" imports: logging and config """
import logging
import logging.config
from biaspeech.utils.config import Config

""" objects:  """
conf = Config() # config
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=conf.LOGLEVEL)

""" MVC """
from biaspeech.models.blocks.block import Block
from biaspeech.models.blocks.arduino import Arduino
from biaspeech.models.blocks.alexa import Alexa
from biaspeech.models.blocks.camera import Camera
from biaspeech.models.blocks.display import Display
from biaspeech.models.blocks.web import Web
from biaspeech.models.blocks.radio import Radio
from biaspeech.models.blocks.tv import Tv
from biaspeech.models.blocks.youtube import Youtube
from biaspeech.models.blocks.netflix import Netflix
from biaspeech.models.blocks.python import Python
from biaspeech.models.blocks.macro import Macro
from biaspeech.models.blocks.default import Default
from biaspeech.utils.lang import Lang

""" objects creation logging and config """
conf = Config() # config
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=conf.LOGLEVEL)

class Bia:
    
    """ constructor """
    def __init__(self, prompt, ai): 
        
        """ attributes """  
        self.conf = Config() # config
        self.logger = logging.getLogger(__name__) # logger        
        self.ai = ai # OpenAI object
        self.prompt = prompt # prompt object
        
        """ initialize """  
        self.lang_utils = Lang() # lang
        
        tp = self.prompt.mia.type # prompt type, like question, alexa etc
        if tp == "alexa": 
            self.block = Alexa(prompt, ai) # block for alexa
        elif tp == "arduino":
            self.block = Arduino(prompt, ai) # block for arduino
        elif tp == "camera":
            self.block = Camera(prompt, ai) # block for camera
        elif tp == "display":
            self.block = Display(prompt, ai) # block for display
        elif tp == "web":
            self.block = Web(prompt, ai) # block for web
        elif tp == "netflix":
            self.block = Netflix(prompt, ai) # block for netflix
        elif tp == "radio":
            self.block = Radio(prompt, ai) # block for radio
        elif tp == "tv":
            self.block = Tv(prompt, ai) # block for tv
        elif tp == "youtube":
            self.block = Youtube(prompt, ai) # block for youtube
        elif tp == "python":
            self.block = Python(prompt, ai) # block for python
        elif tp == "macro":
            self.block = Macro(prompt, ai) # block for macro
        else:
            self.block = Default(prompt, ai) # default block (for other types like question etc)
            
        self.set_file() # set the file

    """ set_file """        
    def set_file(self):
        file = self.prompt.mia.file
        if file == "": # no cache found 
            file = self.block.get_file() # for polymorph blocks like arduino, build the code and store it into the file 
        
        self.prompt.mia.file = file # update mia w the file
                    
    """ get_output """        
    def get_output(self):
        output = self.prompt.mia.output.lower()
        output_en = self.prompt.mia.output_en.lower()
        if output == "": # no cache found 
            self.logger.info("no cache found") # inform the user  
            
            output = self.block.get_output() # get output     
            self.prompt.mia.output = output.lower() # update mia 
            
            output_en = self.lang_utils.get_translate(output, self.prompt.mia.lang, conf.LANG_TO) # back tranlation                                               
            self.prompt.mia.output_en = output_en.lower() # update mia 

        else:
            self.logger.info("cache found") # inform the user  
            
        self.logger.info("output : " + output.lower()) # inform the user   
        self.logger.info("output_en : " + output_en.lower()) # inform the user   
                                      
        return output 

    """ run """        
    def run(self):   
        self.block.run(self.prompt.mia.file) # running 
    
    """ save """        
    def save(self):   
        if str(self.prompt.type) in conf.BACKPROPAGATE: # positive or negative feedback
            self.logger.info("backpropagation") # inform the user 
            self.prompt.mia.backpropagate() # back propagation 
        
        if str(self.prompt.type) == "macro": # macro
            self.logger.info("macro") # inform the user 
            self.prompt.mia.macro() # macro   
                  
        self.prompt.mia.save() # save



# --------------------------
# Models folder
# Class definition
#
# Grandma : CRUD for grandma table
# --------------------------
from websockets.version import commit

""" imports: logging and config """
import logging
import logging.config
from biaspeech.utils.config import Config

""" imports: sqlite """
import sqlite3

""" objects: logging and config """
conf = Config() # config
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=conf.LOGLEVEL)

CONN = sqlite3.connect(conf.DATABASE, check_same_thread=False) # connection to sqlite
CURSOR = CONN.cursor() # cursor

class Grandma:
    
    """ constructor """
    def __init__(self, id, input_en, paraphrase_en, w_distance, datetime): 
        
        self.id = id
        self.input_en = input_en 
        self.paraphrase_en = paraphrase_en 
        self.w_distance = w_distance
        self.datetime = datetime
    
    """ save """
    def save(self):
        sql = """
            insert into grandma ( input_en, paraphrase_en, w_distance, datetime)
            values (?, ?, ?, datetime())
            """
            
        CURSOR.execute(sql, ( self.input_en, self.paraphrase_en, self.w_distance))
        CONN.commit()
        
        self.id = CURSOR.lastrowid 

    """ get_orphan """
    def get_orphan(cls):
        sql = """    
        select distinct input_en as input_en
        from mia
        where input_en not in (select distinct input_en from grandma)
        group by input_en
        """        
        table_rows = CURSOR.execute(sql).fetchall()
          
        return (table_rows)



# --------------------------
# Models folder
# Class definition
#
# Mia : CRUD for mia table
# --------------------------
from websockets.version import commit

""" imports: logging and config """
import logging
import logging.config
from biaspeech.utils.config import Config

""" imports: sqlite """
import sqlite3

""" objects: logging and config """
conf = Config() # config
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=conf.LOGLEVEL)

CONN = sqlite3.connect(conf.DATABASE, check_same_thread=False) # connection to sqlite
CURSOR = CONN.cursor() # cursor

class Mia:
    
    """ constructor """
    def __init__(self, id_cache, type, lang, input, input_en, output, output_en, file,
                 w_alexa, w_arduino, w_camera, w_display, w_web, w_radio, w_tv, w_youtube, w_netflix, w_python, w_macro,
                 w_question, w_positive, w_negative, w_neutral, backpropagation, datetime): 
        
        self.id_cache = id_cache
        self.type = type
        self.lang =lang
        self.input = input
        self.input_en = input_en 
        self.output = output
        self.output_en = output_en
        self.file = file
        self.w_alexa = w_alexa
        self.w_arduino = w_arduino
        self.w_camera = w_camera
        self.w_display = w_display
        self.w_web = w_web
        self.w_radio = w_radio  
        self.w_tv = w_tv
        self.w_youtube = w_youtube
        self.w_netflix = w_netflix
        self.w_python = w_python
        self.w_macro = w_macro
        self.w_question =  w_question
        self.w_positive = w_positive
        self.w_negative = w_negative
        self.w_neutral = w_neutral
        self.backpropagation = backpropagation
        self.datetime = datetime
    
    """ save """
    def save(self):
        sql = """
            insert into mia ( id_cache, type, lang, input, input_en, output, output_en, file,
                 w_alexa, w_arduino, w_camera, w_display, 
                 w_web, w_radio, w_tv, w_youtube, w_netflix, w_python, w_macro,
                 w_question, w_positive, w_negative, w_neutral, backpropagation, datetime)
            values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, datetime())
            """
            
        CURSOR.execute(sql, ( self.id_cache, self.type, self.lang, self.input, self.input_en, 
            self.output, self.output_en, self.file, self.w_alexa, self.w_arduino, self.w_camera, 
            self.w_display, self.w_web, self.w_radio, self.w_tv, self.w_youtube, self.w_netflix, 
            self.w_python, self.w_macro, self.w_question, self.w_positive, self.w_negative, self.w_neutral, self.backpropagation))
            
        CONN.commit()
        self.id = CURSOR.lastrowid 

    """ back_propagate """
    def backpropagate(self):
        sql = """    
        update mia
            set w_positive = ?, w_negative = ?, w_neutral = ?, backpropagation = backpropagation + 1
            where type in ("arduino", "display", "python", "macro", "question")
            and input_en = (
                select input_en from mia
                where type in ("arduino", "display", "python", "macro", "question")
                order by datetime desc
                limit 1 
                )
        """    
            
        CURSOR.execute(sql, ( self.w_positive, self.w_negative, self.w_neutral ))
        CONN.commit
        
    """ macro """
    def macro(self):
        sql = """    
        update mia
            set input = ?, input_en = ?
            where input_en = (
                select input_en from mia
                where type <> 'macro'
                order by datetime desc
                limit 1 
                )
        """    
            
        input = self.input.replace("macro ", "") # cleanup
        input_en = self.input_en.replace("macro ", "") # cleanup
        
        CURSOR.execute(sql, ( input, input_en ))
        CONN.commit
        
    """ create """
    @classmethod 
    def create(cls, id_cache, type, lang, input, input_en, output, output_en, 
        file, w_alexa, w_arduino, w_camera, w_display, w_web, 
        w_radio, w_tv, w_youtube, w_netflix, w_python, w_macro, 
        w_question, w_positive, w_negative, w_neutral, backpropagation, datetime):
        
        new_instance = cls(id_cache, type, lang, input, input_en, output, output_en,
            file, w_alexa, w_arduino, w_camera, w_display, w_web, 
            w_radio, w_tv, w_youtube, w_netflix, w_python, w_macro, 
            w_question, w_positive, w_negative, w_neutral, backpropagation, datetime)
    
        return new_instance
    
    """ create_from_db """
    @classmethod
    def create_from_db(cls, table_row):
        new_instance = cls(table_row[1], table_row[2], table_row[3], table_row[4], table_row[5], table_row[6], table_row[7], 
            table_row[8], table_row[9], table_row[10], table_row[11], table_row[12], table_row[13], 
            table_row[14], table_row[15], table_row[16], table_row[17], table_row[18], table_row[19], table_row[20],
            table_row[21], table_row[22], table_row[23], table_row[24], table_row[25])
        
        new_instance.id = table_row[0]

        return new_instance 

    """ get_table_rows """
    @classmethod
    def get_table_rows(cls):
        sql = """    
            select * from mia
        """
        table_rows = CURSOR.execute(sql).fetchall()

        return [cls.create_from_db(row) for row in table_rows]
    
    """ get_mom """
    @classmethod
    def get_mom(cls, input_en):
        sql = """    
        select id, id, type, lang, 
            input, input_en, output, output_en,
            file, 
            w_alexa, w_arduino, w_camera, w_display, w_web, 
            w_radio, w_tv, w_youtube, w_netflix, w_python, w_macro, 
            w_question, w_positive, w_negative, w_neutral, backpropagation, datetime()
        from mia
        where input_en = ?
        order by w_neutral desc, w_positive desc, datetime desc
        limit 1
        """        
        table_rows = CURSOR.execute(sql, (input_en,)).fetchall()

        return ([cls.create_from_db(row) for row in table_rows])   
    
    """ get_grandma """
    @classmethod
    def get_grandma(cls, input_en):
        sql = """    
        select mia.id, mia.id, type, lang, 
            grandma.input_en, grandma.input_en, output, output_en, 
            file,
            w_alexa, w_arduino, w_camera, w_display, w_web, 
            w_radio, w_tv, w_youtube, w_netflix, w_python, w_macro,
            w_question, w_positive, w_negative, w_neutral, backpropagation, datetime() 
        from mia
        join grandma on mia.input_en = grandma.input_en
        where grandma.paraphrase_en = ?
        order by w_neutral desc, w_positive desc, mia.datetime desc, grandma.w_distance desc
        limit 1
        """        
        table_rows = CURSOR.execute(sql, (input_en,)).fetchall()
          
        return ([cls.create_from_db(row) for row in table_rows])



# --------------------------
# Deploy folder
#
# setup definition
# --------------------------

from setuptools import setup

__project__ = "biaspeech"
__version__ = "0.0.6"
__description__ = "BIA (AI), your AI powered vocal assistant"
__long_description__ = "The module is documented on instructables : https://www.instructables.com/BIA-AI-Your-IA-Powered-Vocal-Assistant-instructabl/"
__packages__ = ["biaspeech", "biaspeech.utils", "biaspeech.controllers", "biaspeech.models.blocks", "biaspeech.models.database", "biaspeech.models.openAI", "biaspeech.models.robot", "biaspeech.models.speech", "biaspeech.data", "biaspeech.views", "biaspeech.bin"]
__author__ = "TylerDDDD"
__author_email__ = "makertylerdddd@gmail.com"
__classifiers__ = [
	"Development Status :: 3 - Alpha", 
	"Intended Audience :: Education", 
	"Programming Language :: Python :: 3",
]
__readme__ = "README.md"
__keywords__ = ["robot", "otto", "ai", "vocal", "assistant",]
__requires__ = ["nltk", "openai", "websockets", "argostranslate", "py3langid", "pyautogui", "pyaudio", "speechrecognition", "pyttsx3", "parrot", "opencv-python", "customtkinter", "pyduinocli"]

setup(
    name = __project__,
    version = __version__,
    description = __description__,
    long_description = __long_description__,
    packages = __packages__,
    author = __author__,
    author_email = __author_email__,
    classifiers = __classifiers__,
    readme = __readme__,
    keywords = __keywords__,
    install_requires = __requires__,
    package_data={'': ['bin/*', 'views/frame0/*', 'utils/*cfg', 'data/*', 'data/arduino/*', 'data/camera/*', 'data/logs/*', 'data/python/*', 'README.md', 'LICENSE']},
    include_package_data=True,
)




# --------------------------
# Models folder
# Class definition
#
# Speech2Text : define the speech to text model
# --------------------------

""" imports: logging and config """
import logging
import logging.config
from biaspeech.utils.config import Config
import os

""" imports: speech_recognition """
import speech_recognition as sr 

""" objects:  """
conf = Config() # config
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=conf.LOGLEVEL)

class Speech2Text:
    
    """ constructor """
    def __init__(self, language): 

        """ util objects """
        self.conf = Config() # config
        self.logger = logging.getLogger(__name__) # logger

        """ attributes """
        self.language = language # language
        self.input = "" # input text
        
    """ get_input """        
    def get_input(self):
        return self.input
                                  
    """ do_listen """        
    def do_listen(self):
        recognizer = sr.Recognizer() # initialize the recognizer    
        r = sr.Recognizer()
        
        try:
            if (conf.OS == "macos"): # macos                                                                                   
                with sr.Microphone() as source: # microphone                                                           
                    recorded_audio = recognizer.listen(source) # listen
                    #recognizer.adjust_for_ambient_noise(source, duration=conf.ADJUST)
                    #recorded_audio = recognizer.listen(source,timeout=0,phrase_time_limit=conf.TIMEOUT)
                    text = recognizer.recognize_google(recorded_audio, language=self.language) # w language
                    
            else: # raspberry
                stream = os.popen(conf.RECORD) # stream
                audio = sr.AudioFile(conf.RECORDING) # recording

                with audio as source:
                    audio = r.record(source) # recording                 
                    text = r.recognize_google(audio, language = self.language) # w language
            
            self.input = format(text) # format
                                                       
        except sr.UnknownValueError:
            self.logger.warning(conf.ERR_A) # inform the user
        except sr.RequestError:
            self.logger.warning(conf.ERR_B) # inform the user
        except Exception as ex:
            self.logger.warning(conf.ERR_C) # inform the user
